//
// Copyright (C) 2003 Trevor Hogan
// Copyright (C) 2007 Joel Yliluoma
//

#ifndef ATOM_H
 #define ATOM_H

#include <stdint.h>
#include "autoptr.hh"

#include <vector>
#include <map>

class CAtomInt;
class CAtomString;
class CAtomList;
class CAtomDict;
class CAtom: public ptrable
{
public:
    CAtom() : ptrable() { }
    virtual ~CAtom( ) { }
    
    const CAtomInt* get_int() const;
    const CAtomList* get_list() const;
    const CAtomDict* get_dict() const;
    const CAtomString* get_str() const;

    virtual autoptr<CAtom> clone() const = 0;
    virtual std::string bencode( ) const = 0;

    unsigned EncodedLength( ) const { return bencode().size(); }
};

class CAtomInt : public CAtom
{
public:
    CAtomInt( ) : CAtom(), m_iInt(0) { }
    CAtomInt( int_fast64_t iInt ) : CAtom(), m_iInt(iInt) { }
    CAtomInt( const CAtomInt &c ) : CAtom(), m_iInt(c.m_iInt) { }
    virtual ~CAtomInt( ) { }

    virtual autoptr<CAtom> clone() const { return new CAtomInt(*this); }

    virtual std::string bencode( ) const;

    int_fast64_t getValue( ) const { return m_iInt; }
    void setValue( int_fast64_t iInt ) { m_iInt = iInt; }

private:
    int_fast64_t m_iInt;
};

class CAtomString : public CAtom
{
public:
    CAtomString( ) : CAtom(), m_strString() { }
    CAtomString( std::string strString ) : CAtom(), m_strString(strString) { }
    CAtomString( const char* strString ) : CAtom(), m_strString(strString) { }
    CAtomString( const CAtomString &c ) : CAtom(), m_strString(c.m_strString) { }
    virtual ~CAtomString( ) { }

    virtual autoptr<CAtom> clone() const { return new CAtomString(*this); }

    virtual std::string bencode( ) const;

    const std::string& getValue( ) const { return m_strString; }
    void setValue( const std::string& strString ) { m_strString = strString; }

private:
    std::string m_strString;
};

class CAtomRawBencode : public CAtom
{
public:
    explicit CAtomRawBencode(const std::string& s) : CAtom(), data(s) { }
    virtual ~CAtomRawBencode( ) { }

    virtual autoptr<CAtom> clone() const { return new CAtomRawBencode(*this); }
    virtual std::string bencode( ) const { return data; }

private:
    std::string data;
};

class CAtomList : public CAtom
{
public:
    CAtomList( ) : CAtom(), m_vecList() { }
    CAtomList( const std::vector<autoptr<CAtom> > vecList ) : CAtom(), m_vecList()
        { setValue(vecList); }
    CAtomList( const CAtomList &c ) : CAtom(), m_vecList() { setValue(c.m_vecList); }
    virtual ~CAtomList( ) { }

    virtual autoptr<CAtom> clone() const { return new CAtomList(*this); }

    virtual std::string bencode( ) const;

    virtual bool empty( ) const { return m_vecList.empty(); }
    virtual void clear( ) { m_vecList.clear(); }

    virtual void Randomize( );

    const std::vector<autoptr<CAtom> >& getValue( ) const { return m_vecList; }
    
    void setValue(const std::vector<autoptr<CAtom> >& vecList );

    void delItem( autoptr<CAtom> atmItem );
    void addItem( autoptr<CAtom> atmItem );
    
    void chop_to(unsigned newsize) { if(newsize < size()) m_vecList.resize(newsize); }
    
    unsigned size() const { return m_vecList.size(); }
    autoptr<CAtom> operator[] (unsigned ind) const
        { return ind < size() ? &*m_vecList[ind] : (CAtom*)0; }

private:
    std::vector<autoptr<CAtom> > m_vecList;
};

class CAtomDict : public CAtom
{
public:
    typedef std::map<std::string, autoptr<CAtom> > dict_t;
    
    CAtomDict( ) : CAtom(), m_mapDict() { }
    CAtomDict( const CAtomDict &c ) : CAtom(), m_mapDict()
        { setValue(c.m_mapDict); }
    virtual ~CAtomDict( ) { }

    virtual autoptr<CAtom> clone() const { return new CAtomDict(*this); }

    virtual std::string bencode( ) const;

    virtual bool empty( ) const { return m_mapDict.empty(); }
    virtual void clear( ) { m_mapDict.clear(); }

    const dict_t& getValue( ) const { return m_mapDict; }
    
    void setValue(const dict_t& mapDict );

    autoptr<CAtom> getItem(const std::string& strKey, autoptr<CAtom> ret=0) const;

    CAtomInt getItemInt(const std::string& strKey, CAtomInt def=CAtomInt()) const;
    CAtomString getItemStr(const std::string& strKey, CAtomString def=CAtomString()) const;
    CAtomList getItemList(const std::string& strKey, CAtomList def=CAtomList()) const;
    CAtomDict getItemDict(const std::string& strKey, CAtomDict def=CAtomDict()) const;
    
    void delItem( const std::string& strKey );
    void setItem( const std::string& strKey, autoptr<CAtom> pValue );
    
    autoptr<CAtom> operator[] (const std::string& key) const { return getItem(key); }

private:
    dict_t m_mapDict;
};

#endif
