/*

    Using a threading server with boost::thread.
    
    This idea was cancelled when it was found that a mechanism
    like this becomes needlessly complicated with all the mutex
    stuff, and actually provides no extra performance to
    compensate it.

*/
#include <stdint.h>
#include <string>
#include <vector>
#include <sstream>
#include <map>
#include <set>
#include <csignal>

#include <unistd.h>
#include <glob.h>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/thread.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/bind.hpp>

#include "hash.hh"
#include "sha1.hh"
#include "socket.hh"
#include "httpfun.hh"
#include "bdecode.hh"
#include "autoptr.hh"
#include "stringfun.hh"

struct AutoCloser { AutoCloser(int s) : sock(s) { }
                    ~AutoCloser() { close(sock); }
                    int sock; };
template<typename F>
struct AutoUnlocker { AutoUnlocker(F& f) : ff(f) { }
                      ~AutoUnlocker() { ff.do_unlock(); }
                      F& ff; };

static bool is_local_ip(const std::string& s)
{
    if(s.compare(0,3,"10.") == 0) return true;
    if(s.compare(0,4,"127.") == 0) return true;
    if(s.compare(0,8,"192.168.") == 0) return true;
    if(s.compare(0,8,"169.254.") == 0) return true;
    if(s.compare(0,7,"172.16.") == 0) return true;
    if(s.compare(0,7,"172.17.") == 0) return true;
    if(s.compare(0,7,"172.18.") == 0) return true;
    if(s.compare(0,7,"172.19.") == 0) return true;
    if(s.compare(0,6,"172.2.") == 0) return true;
    if(s.compare(0,7,"172.30.") == 0) return true;
    if(s.compare(0,7,"172.31.") == 0) return true;
    return false;
}

static const std::string get_forwarded_ip(const std::string& ip,
    const std::map<std::string, std::string>& headers)
{
    std::map<std::string, std::string>::const_iterator i;
    i = headers.find("http_x_forwarded_for");
    if(i != headers.end())
    {
        const std::string& forw = i->second;
        std::string::size_type p = forw.find(',');
        if(p == forw.npos) return forw;
        std::string res = forw.substr(0, p);
        if(!is_local_ip(res)) return res;
        return forw.substr(p);
    }
    
    i = headers.find("http_client_ip");
    if(i != headers.end()) return i->second;

    i = headers.find("http_via");
    if(i != headers.end())
    {
        const std::string& via = i->second;
        std::string::size_type p = via.find(" for ");
        if(p != via.npos) return via.substr(p + 5);
    }
    
    i = headers.find("http_from");
    if(i != headers.end()) return i->second;
    
    return ip;
}

struct HTTPresponse // for throwing
{
    HTTPresponse(int cde, const std::string& ctype,
                          const std::string& msg)
        : code(cde),contenttype(ctype),content(msg),extraheaders()
    {
    }
    void AddHeader(const std::string& hdr, const std::string& value)
        { extraheaders[hdr]=value; }
    
    int code;
    std::string contenttype;
    std::string content;
    std::map<std::string, std::string> extraheaders;
    
    const std::string ConvertToResponse() const
    {
        std::string resptype = "RESPONSE";
        switch(code)
        {
            case 200: resptype = "OK"; break;
            case 301:
            case 302: resptype = "ALMOST RIGHT"; break;
            case 400: resptype = "BAD REQUEST"; break;
            case 404: resptype = "NOT FOUND"; break;
            case 500: resptype = "OH THE PAIN"; break;
        }
        
        std::stringstream resptext;
        resptext << "HTTP/1.1 " << code << " " << resptype << "\r\n"
                    "Content-type: " << contenttype << "\r\n"
                    "Content-length: " << content.size() << "\r\n";
        for(std::map<std::string, std::string>::const_iterator
            i = extraheaders.begin();
            i != extraheaders.end();
            ++i)
        {
            resptext << i->first << ": " << i->second << "\r\n";
        }
        resptext << "\r\n" << content;
        return resptext.str();
    }
};

class Tracker
{
private:
    struct Peer
    {
        std::string peer_id;
        std::string addr_ip;
        int         addr_port;
        
        std::string key;
        // if the peer's key changes, no changes must be done to this struct.
        
        int_fast64_t left; // how much to still transfer
        
        // nat, useragents
        
        time_t first_seen;
        time_t last_seen;
        
        Peer()
            : peer_id(),
              addr_ip(),
              addr_port(0),
              key(),
              left(0),
              first_seen(0),
              last_seen(0) { }
    };

    struct File
    {
        enum FileClass
        {
            Allowed,
            Obsolete,
            Unknown
        } fclass;
        
        std::string infohash;
        
        std::string file;          // .torrent
        std::string name;          // contained file
        std::string path;          // path/.torrent
        int_fast64_t length;       // of contained file
        int_fast64_t torrentsize;
        unsigned numfiles;
        
        hash_map<std::string /* peer_id */,
                 Peer> peers;

        uint_fast64_t completed; // total completed downloads
        
        /* shared_ptr resolves the problem that mutexes cannot be copied */
        boost::mutex mut;
        boost::mutex::scoped_lock lck;
    
        void do_lock()   { lck.lock(); }
        void do_unlock() { lck.unlock(); }
        
    public:
        File()
            : fclass(Unknown),
              infohash(),
              file(),
              name(),
              path(),
              length(),
              torrentsize(),
              numfiles(),
              peers(),
              completed(),
              mut(), lck(mut,false) {}
        File(const File& f)
            : fclass(f.fclass),
              infohash(f.infohash),
              file(f.file),
              name(f.name),
              path(f.path),
              length(f.length),
              torrentsize(f.torrentsize),
              numfiles(f.numfiles),
              peers(f.peers),
              completed(f.completed),
              mut(), lck(mut,false) { }
        File& operator= (const File& f)
        {
            fclass = (f.fclass);
            infohash = (f.infohash);
            file = (f.file);
            name = (f.name);
            path = (f.path);
            length = (f.length);
            torrentsize = (f.torrentsize);
            numfiles = (f.numfiles);
            peers = (f.peers);
            completed = (f.completed);
        }
        
        void CleanupPeers()
        {
            // TODO: remove peers who haven't talked for a long time
            //       and keep the number of them low anyway.
        }
    };

private:
    hash_map<std::string /* info hash */,
             File> files;
    
    int listen_port;
    std::vector<std::string> accepted_dirs;
    std::string redirect_url;
    std::string dstate_file;
    unsigned dstate_write_interval;
    unsigned parse_dir_interval;
    unsigned reannounce_interval;
    
    boost::thread_group threads;
    bool terminate;
    boost::mutex files_mutex;

public:
    Tracker()
        : files(),
          listen_port(6969/* * 5*/),
          accepted_dirs(),
          redirect_url("http://tasvideos.org/"),
          dstate_file("dstate"),
          dstate_write_interval(30),
          parse_dir_interval(60),
          reannounce_interval(30*60),
          
          threads(),
          terminate(false),
          files_mutex()
    {
        accepted_dirs.push_back("/WWW/torrents");
        
        LoadState();
        ReloadTorrentRepository();
        
        MainLoop();
    }

    void MainLoop()
    {
        boost::thread* s = threads.create_thread(boost::bind(&Tracker::ServerLoop,         this));
        boost::thread* w = threads.create_thread(boost::bind(&Tracker::WriteStateLoop,     this));
        boost::thread* r = threads.create_thread(boost::bind(&Tracker::ReloaderLoop,       this));
        s->join();
        r->join();
        w->join();
    }
private:
    class LockAllFiles: private boost::noncopyable
    {
    public:
        LockAllFiles(Tracker& t)
            : tracker(t), lock(t.files_mutex)
        {
            for(hash_map<std::string, File>::iterator
                i = tracker.files.begin(); i != tracker.files.end(); ++i)
            {
                // Lock and unlock them in succession to ensure
                // that nobody is using them. We don't need to worry
                // about someone locking them after this, because the
                // only way locks can be acquired is through files_mutex,
                // which is already locked.
                
                i->second.do_lock();
                i->second.do_unlock();
            }
        }
    private:
        Tracker& tracker;
        boost::mutex::scoped_lock lock;
    };

    void LoadState() // loads "dstate"
    {
        boost::scoped_ptr<LockAllFiles> AllFilesLock ( new LockAllFiles(*this) );
        // TODO

        std::string Dstate = LoadFileContents(dstate_file);
        autoptr<CAtom> decoded = BDecode(Dstate);
        if(!decoded) return;
        const CAtomDict* info = decoded->get_dict();
        if(!info) return;
        
        time_t t = time(NULL);
        
        CAtomDict peers_dict     = info->getItemDict("peers");
        CAtomDict completed_dict = info->getItemDict("completed");
        CAtomDict allowed_dict   = info->getItemDict("allowed");

        const std::map<std::string, autoptr<CAtom> >& cdict = completed_dict.getValue();
        for(std::map<std::string, autoptr<CAtom> >::const_iterator
            i = cdict.begin(); i != cdict.end(); ++i)
        {
            if(!i->second) continue;
            const CAtomInt* val = i->second->get_int(); if(!val) continue;
            File& file = files[i->first];
            file.fclass    = File::Obsolete;
            file.infohash  = i->first;
            file.completed = val->getValue();
        }

        const std::map<std::string, autoptr<CAtom> >& adict = allowed_dict.getValue();
        for(std::map<std::string, autoptr<CAtom> >::const_iterator
            i = adict.begin(); i != adict.end(); ++i)
        {
            if(!i->second) continue;
            const CAtomDict* dic = i->second->get_dict(); if(!dic) continue;
            File& file = files[i->first];
            file.infohash  = i->first;
            file.name      = dic->getItemStr("name").getValue();
            file.file      = dic->getItemStr("file", file.name + ".torrent").getValue();
            file.path      = dic->getItemStr("path", file.file).getValue();
            file.numfiles  = dic->getItemInt("numfiles", 1).getValue();
            file.length    = dic->getItemInt("length", 0).getValue();
            file.fclass    = File::Allowed;
            
            CAtomDict peers_here = peers_dict.getItemDict(i->first);
            const std::map<std::string, autoptr<CAtom> >& pdict = peers_here.getValue();
            for(std::map<std::string, autoptr<CAtom> >::const_iterator
                p = pdict.begin(); p != pdict.end(); ++p)
            {
                if(!p->second) continue;
                const CAtomDict* pee = p->second->get_dict(); if(!pee) continue;
                Peer& peer = file.peers[p->first];
                peer.peer_id = p->first,
                peer.addr_ip   = pee->getItemStr("ip").getValue();
                peer.key       = pee->getItemStr("key").getValue();
                peer.left      = pee->getItemInt("left").getValue();
                // nat not used
                peer.addr_port = pee->getItemInt("port").getValue();
                // useragents ignored
                peer.first_seen = t;
                peer.last_seen  = t;
            }
        }
    }
    void WriteState() // writes "dstate"
    {
        boost::scoped_ptr<LockAllFiles> AllFilesLock ( new LockAllFiles(*this) );
        // TODO
    }
    
    File& FindFileAndLock(const std::string& info_hash)
    {
        boost::mutex::scoped_lock lock(files_mutex);
        
        hash_map<std::string, File>::iterator i = files.find(info_hash);
        if(i == files.end()) throw false;
        i->second.do_lock();
        
        // will relinquish global lock here, leaving the singular lock active.
        return i->second;
    }
    
    void ReloadTorrentRepository() // rereads the accepted dirs
    {
        std::set<std::string> found_infohashes;
        
        for(unsigned a=0; a<accepted_dirs.size(); ++a)
        {
            namespace fs = boost::filesystem;
            
            fs::path fpath(fs::system_complete(fs::path(accepted_dirs[a], fs::native)));
            if(!fs::exists(fpath)) continue;
            if(!fs::is_directory(fpath)) continue;
            fs::directory_iterator end_i;
            for(fs::directory_iterator dir_i(fpath); dir_i != end_i; ++dir_i)
                if(!is_directory(*dir_i))
                {
                    /*
                    fprintf(stderr, "s(%s)rn(%s)rd(%s)leaf(%s)\n",
                        dir_i->string().c_str(),
                        dir_i->root_name().c_str(),
                        dir_i->root_directory().c_str(),
                        dir_i->leaf().c_str());
                    */
                    std::string hash = TorrentRepositoryUpdate(
                        dir_i->string(),
                        dir_i->leaf());
                    if(!hash.empty())
                    {
                        //fprintf(stderr, "ok '%s'\n", hash.c_str());
                        found_infohashes.insert(hash);
                    }
                }
        }
        
        /* Lock files */
        boost::scoped_ptr<LockAllFiles> AllFilesLock ( new LockAllFiles(*this) );
        
        // Now check the "files" list; for every "Allowed" hash that
        // was _not_ found in found_infohashes, set type as "Obsolete".
        for(hash_map<std::string, File>::iterator
            i = files.begin(); i != files.end(); ++i)
        {
            if(found_infohashes.find(i->first) == found_infohashes.end()
            && i->second.fclass == File::Allowed)
            {
                fprintf(stderr, "Obsoleting torrent %s (%s)\n",
                    i->second.name.c_str(),
                    i->first.c_str());
                i->second.fclass = File::Obsolete;
            }
        }
    }
    
    std::string TorrentRepositoryUpdate(
        const std::string& pathname,
        const std::string& filename)
    {
        /* Read a torrent file, verify its validity, and gather
         * the relevant data into the Files struct.
         */
        
        try {
        std::string Torrent = LoadFileContents(pathname);
        autoptr<CAtom> decoded = BDecode(Torrent);
        if(const CAtomDict* dict = decoded->get_dict())
        {
            //fprintf(stderr, "dict: %s\n", dict->bencode().c_str());
            
            if(dict->getItemStr("announce").getValue().empty()
            && dict->getItemList("announce-list").empty()) throw __LINE__;
            
            CAtomDict infodict = dict->getItemDict("info");
            
            if(infodict.getItemStr("pieces").getValue().size() % 20) throw __LINE__;
            if(infodict.getItemInt("piece length").getValue() <= 0) throw __LINE__;
            const std::string name = infodict.getItemStr("name").getValue();
            if(name.empty() || name.find('/') != name.npos
                            || name.find('\\') != name.npos
                            || name[0]=='.'
                            || name[0]=='~') throw __LINE__;
            
            //fprintf(stderr, "has%s files\n", !infodict.getItem("files") ? "no" : "");
            //fprintf(stderr, "has%s length\n", !infodict.getItem("length") ? "no" : "");
            
            if(!infodict.getItem("files") == !infodict.getItem("length")) throw __LINE__;
            
            int_fast64_t length = infodict.getItemInt("length").getValue();
            if(length < 0) throw __LINE__;
            CAtomList files = infodict.getItemList("files");
            for(unsigned a=0; a<files.size(); ++a)
            {
                const CAtomDict* file = files[a]->get_dict();
                if(!file) throw __LINE__;
                length += file->getItemInt("length").getValue();
            }
            
            std::string info_hash = SHA1(infodict.bencode());
            
            //for(unsigned a=0; a<info_hash.size(); ++a)
            //    info_hash[a] = 32 + (info_hash[a]&63);
            
            boost::scoped_ptr<LockAllFiles> AllFilesLock ( new LockAllFiles(*this) );
            
            File& file = this->files[info_hash];
            
            file.infohash = info_hash;
            file.length = length;
            file.numfiles = files.empty() ? 1 : files.size();
            file.name = name;
            file.path = pathname;
            file.file = filename;
            file.torrentsize = Torrent.size();
            
            if(file.fclass != File::Allowed)
            {
                file.fclass = File::Allowed;
                /*
                fprintf(stderr, "Registered len(%u)num(%u) name(%s)path(%s)file(%s)\n",
                    (unsigned)file.length,
                    (unsigned)file.numfiles,
                    file.name.c_str(),
                    file.path.c_str(),
                    file.file.c_str()); */
            }
            return info_hash;
        }
        } catch(int l) { fprintf(stderr, "torrent rejected at line %d\n", l); }
        return std::string();
    }
    
    void WriteStateLoop()
    {
        while(!terminate) { sleep(dstate_write_interval); WriteState(); }
    }
    
    void ReloaderLoop()
    {
        while(!terminate) { sleep(parse_dir_interval); ReloadTorrentRepository(); }
    }
    
    void ServerLoop() // spawn a new thread for each incoming connection
    {
        int plug = MakeServerSocket(listen_port, true);
        if(plug < 0)
        {
            terminate = true;
            return;
        }
        
        const AutoCloser AutoClose(plug);
        
        while(!terminate)
        {
            int sock = SockAccept(plug, true);
            if(sock < 0) break;
            
            //fprintf(stderr, "Creating thread to handle sock %d\n", sock);
            
            std::string ip = GetLastPeerIp();
            threads.create_thread(boost::bind(&Tracker::HandleRequest, this, sock, ip));

            //fprintf(stderr, "Created thread to handle sock %d\n", sock);
        }
    }
    
    void HandleRequest(int sock, const std::string& ip)
    {
        /* Ensure the socket will be closed when this function exits,
         * even if an uncaught exception is thrown
         */
        const AutoCloser AutoClose(sock);
        
        //fprintf(stderr, "HandleReq %d from '%s'\n", sock, ip.c_str());
        
        FILE* fp = fdopen(sock, "rb+");
        char Buf[4096] = "";
        char method[4096] = "", reqstr[4096] = "", proto[4096] = "";
        
        fgets(Buf, sizeof(Buf), fp);
        { char* c = strchr(Buf,'\r'); if(c) *c=0; }
        { char* c = strchr(Buf,'\n'); if(c) *c=0; }
        const std::string request_line = Buf;
        sscanf(Buf, "%4096s %4096s %4096s", method, reqstr, proto);
        
        std::string req(reqstr);
        std::multimap<std::string, std::string> queryparams;
        SplitDecodeQuery(req, queryparams);
        
        std::string remote_ip = ip;
        
        std::map<std::string, std::string> headers;
        while(fgets(Buf, sizeof(Buf), fp))
        {
            { char* c = strchr(Buf,'\r'); if(c) *c=0; }
            { char* c = strchr(Buf,'\n'); if(c) *c=0; }
            
            if(!*Buf) break;
            //fprintf(stderr, "hdr line '%s'\n", Buf);
            
            // It is a header.
            char* colon = strchr(Buf, ':');
            if(colon)
            {
                std::string hdrname = LCase(Trim(std::string(Buf,colon)));
                for(unsigned a=0; a<hdrname.size(); ++a)
                    if(hdrname[a]=='-') hdrname[a]='_';
                while(*++colon == ' ');
                std::string hdrvalue= Trim(colon);
                headers[hdrname] = hdrvalue;
            }
        }
        
        remote_ip = get_forwarded_ip(ip, headers);
        HTTPresponse resp = HandleReq(method, req, queryparams, headers, remote_ip);
        
        //fprintf(stderr, "Got response: '%s' (%u bytes)\n", response.c_str(), response.size());
        
        char datetime[128];
        struct tm tm; { time_t t = time(NULL); localtime_r(&t, &tm); }
        strftime(datetime,sizeof(datetime), "%d/%b/%Y:%H:%M:%S", &tm);
        
        printf("%s - - [%s] \"%s\" %d %lu \"-\" \"%s\"\n",
            ip.c_str(),
            datetime,
            request_line.c_str(),
            resp.code,
            (unsigned long)resp.content.size(),
            headers["user_agent"].c_str()
        );
            
        /* TODO: handle sigpipes nicely */
        const std::string response = resp.ConvertToResponse();
        fwrite(response.data(), 1, response.size(), fp);
        fflush(fp);
    }
    
    const HTTPresponse
        HandleReq(const std::string& method,
                  const std::string& req,
                  const std::multimap<std::string, std::string>& queryparams,
                  const std::map<std::string, std::string>& headers,
                  const std::string& remote_ip)
    {
        try {
            if(method != "GET") throw HTTPresponse(400, "text/plain",
                "Your browser sent a request we could not understand.");
            
            if(req == "/announce") { ReqAnnounce(queryparams,headers, remote_ip); }
            else if(req == "/" || req == "/index.html") { ReqIndex(queryparams,headers); }
            else if(req == "/scrape")   { ReqScrape(queryparams,headers); }
            else if(req == "/file")     { ReqFile(queryparams,headers); }
            else { ReqOther(); }
        }
        catch(const HTTPresponse& resp)
        {
            return resp;
        }
        return HTTPresponse(500, "text/plain", "Internal server error.");
    }
    
    void ReqScrape( // req for torrent info
        const std::multimap<std::string, std::string>& queryparams,
        const std::map<std::string, std::string>& headers)
    {
        // A number of 'info_hash'es can be given.
        // If none is given, all allowed files should be listed.
        std::multimap<std::string, std::string>::const_iterator
            i = queryparams.find("info_hash");
        
        autoptr<CAtomDict> result = new CAtomDict;
        
        if(i == queryparams.end())
        {
            boost::mutex::scoped_lock lock(files_mutex);
            for(hash_map<std::string, File>::iterator
                i = files.begin(); i != files.end(); ++i)
            {
                if(i->second.fclass == File::Allowed)
                    ScrapeAddFile(*result, i->second);
            }
        }
        else do
        {
            try {
                File& file = FindFileAndLock(i->second);
                AutoUnlocker<File> unlocker(file);
                ScrapeAddFile(*result, file);
            } catch(bool) { }
        } while(++i != queryparams.end() && i->first == "info_hash");
        
        autoptr<CAtomDict> res2 = new CAtomDict;
        res2->setItem("files", &*result);
        
        throw HTTPresponse(200, "application/octet-stream", res2->bencode());
    }
    
    autoptr<CAtomDict> ScrapeCreateInfoDict(const File& file)
    {
        unsigned n_complete   = 0;
        unsigned n_incomplete = 0;
        
        for(hash_map<std::string, Peer>::const_iterator
            i = file.peers.begin(); i != file.peers.end(); ++i)
        {
            if(i->second.left) ++n_incomplete; else ++n_complete;
        }
        
        autoptr<CAtomDict> filedata = new CAtomDict;
        filedata->setItem("downloaded", new CAtomInt(file.completed));
        filedata->setItem("incomplete", new CAtomInt(n_incomplete));
        filedata->setItem("complete",   new CAtomInt(n_complete));
        filedata->setItem("name", new CAtomString(file.name));
        return filedata;
    }
    
    void ScrapeAddFile(CAtomDict& target, const File& file)
    {
        autoptr<CAtomDict> filedata = ScrapeCreateInfoDict(file);
        target.setItem(file.infohash, new CAtomRawBencode(filedata->bencode()));
    }
    
    const std::string GetParam(const std::multimap<std::string, std::string>& qp,
        const std::string& name, const std::string& def = "")
    {
        std::multimap<std::string, std::string>::const_iterator i = qp.find(name);
        if(i == qp.end()) return def;
        return i->second;
    }
    
    void ReqAnnounce( // req for peer info
        const std::multimap<std::string, std::string>& qp,
        const std::map<std::string, std::string>& headers,
        const std::string& remote_ip)
    {
        std::string inp_info_hash = GetParam(qp, "info_hash");
        std::string inp_peer_id   = GetParam(qp, "peer_id");
        std::string inp_key       = GetParam(qp, "key");
        std::string inp_ip        = remote_ip;
        long         inp_port     = strtol(GetParam(qp, "port", "0").c_str(), 0, 10);
        int_fast64_t inp_left     = strtoll(GetParam(qp, "left", "0").c_str(), 0, 10);
        long         numwant      = strtol(GetParam(qp, "numwant", "50").c_str(), 0, 10);
        std::string event         = GetParam(qp, "event");
        
        int return_type = 0;
        
        if(GetParam(qp, "no_peer_id", "0") != "0") return_type = 1;
        if(GetParam(qp, "compact","0")     != "0") return_type = 2;
        
        if(numwant > 50) numwant = 50;
        
        if(inp_info_hash.empty())   throw HTTPresponse(400, "text/plain", "needs info_hash");
        if(inp_peer_id.size() != 20)throw HTTPresponse(400, "text/plain", "id not of length 20");
        if(event != "started" && event != "completed"
        && event != "stopped" && event != "snooped"
        && event != "") 
            throw HTTPresponse(400, "text/plain", "invalid event");
        if(inp_port < 0 || inp_port > 65535) 
            throw HTTPresponse(400, "text/plain", "invalid port");
        if(inp_left < 0)
            throw HTTPresponse(400, "text/plain", "invalid amount left");
        
        try
        {
            File& file = FindFileAndLock(inp_info_hash);
            AutoUnlocker<File> unlocker(file);
            
            file.CleanupPeers();
            
            Peer& peer = file.peers[inp_peer_id];
            
            bool auth = peer.peer_id.empty()
                     || peer.key     == inp_key
                     || peer.addr_ip == inp_ip;

            if(event == "stopped")
            {
                if(auth)
                    file.peers.erase(inp_peer_id);
            }
            else if(peer.peer_id.empty())
            {
                peer.peer_id   = inp_peer_id;
                peer.addr_ip   = inp_ip;
                peer.addr_port = inp_port;
                peer.key       = inp_key;
                peer.left      = inp_left;
                peer.first_seen= time(NULL);
                
                if(event == "completed")
                    file.completed += 1;
            }
            else
            {
                if(auth) 
                {
                    // update stats
                    if(peer.left && !inp_left)
                        file.completed += 1;

                    if(peer.left) peer.left = inp_left;
                    peer.last_seen = time(NULL);
                }
            }

            autoptr<CAtomList> peerslist = new CAtomList;
            // each peers has "ip"(str), "peer id"(str), "port"(int)
            
            for(hash_map<std::string, Peer>::const_iterator
                i = file.peers.begin(); i != file.peers.end(); ++i)
            {
                const Peer& peer = i->second;
                if(peer.peer_id == inp_peer_id) continue;
            
                switch(return_type)
                {
                    case 0: // default
                    {
                        autoptr<CAtomDict> peerdata = new CAtomDict;
                        peerdata->setItem("port",    new CAtomInt(peer.addr_port));
                        peerdata->setItem("ip",      new CAtomString(peer.addr_ip));
                        peerdata->setItem("peer id", new CAtomString(peer.peer_id));
                        peerslist->addItem(&*peerdata);
                        break;
                    }
                    case 1: // default, except peer id is not displayed
                    {
                        autoptr<CAtomDict> peerdata = new CAtomDict;
                        peerdata->setItem("port",    new CAtomInt(peer.addr_port));
                        peerdata->setItem("ip",      new CAtomString(peer.addr_ip));
                        peerslist->addItem(&*peerdata);
                        break;
                    }
                    case 2: // compact
                    {
                        std::string raw;
                        unsigned a=0,b=0,c=0,d=0;
                        sscanf(peer.addr_ip.c_str(), "%d.%d.%d.%d",&a,&b,&c,&d);
                        raw += (char)a;
                        raw += (char)b;
                        raw += (char)c;
                        raw += (char)d;
                        raw += (char)(peer.addr_port >> 8);
                        raw += (char)(peer.addr_port & 255);
                        peerslist->addItem(new CAtomRawBencode(raw));
                        break;
                    }
                }
            }
            
            peerslist->Randomize();
            peerslist->chop_to(numwant);
            
            if(return_type == 2)
            {
                std::string raw;
                for(unsigned a=0; a<peerslist->size(); ++a)
                    raw += (*peerslist)[a]->bencode();
                peerslist = (CAtomList*)(new CAtomString(raw));
            }
            
            autoptr<CAtomDict> result = ScrapeCreateInfoDict(file);
            // ^for complete, incomplete etc
            result->setItem("interval", new CAtomInt(reannounce_interval));
            result->setItem("peers",    &*peerslist);
            
            throw HTTPresponse(200, "application/octet-stream",
              result->bencode());
        }
        catch(bool)
        {
            throw HTTPresponse(200, "text/plain",
                "d14:failure reason"
                 "63:Requested download is not authorized for use with this tracker."
                "e");
        }
    }
    
    void ReqFile( // req for a torrent file
        const std::multimap<std::string, std::string>& queryparams,
        const std::map<std::string, std::string>& headers)
    {
        // info_hash denotes the file to download
        try
        {
            std::multimap<std::string, std::string>::const_iterator
                i = queryparams.find("info_hash");
            if(i == queryparams.end()) throw false;

            File& file = FindFileAndLock(i->second);
            AutoUnlocker<File> unlocker(file);
            
            HTTPresponse res(200, "application/x-bittorrent",
                LoadFileContents(file.path));
            res.AddHeader("Content-Disposition", "attachment; filename="+file.file);
            
            throw res;
        }
        catch(bool)
        {
            HTTPresponse res(404, "text/plain", "Your file is not known to this tracker.");
        }
    }
    
    void ReqIndex( // req for html torrent list
        const std::multimap<std::string, std::string>& queryparams,
        const std::map<std::string, std::string>& headers)
    {
        HTTPresponse res(302, "text/html",
            "Content is at <a href='"+htmlencode(redirect_url)+"'>"
           + htmlencode(redirect_url) + "</a>.");
        res.AddHeader("Location", redirect_url);
        res.AddHeader("Server",     "BQTracker-"VERSION);
        res.AddHeader("Vary",       "Accept-Encoding");
        res.AddHeader("Connection", "close");
        throw res;
    }

    void ReqOther() // req for anything else; respond with an error
    {
        throw HTTPresponse(404, "text/plain", "Data not found at given address.");
    }
    
    const std::string LoadFileContents(const std::string& pathname)
    {
        FILE* fp = fopen(pathname.c_str(), "rb");
        std::string result;
        if(fp)
        {
            for(;;)
            {
                char Buf[8192];
                int n = fread(Buf, 1, sizeof(Buf), fp);
                if(n == 0 && ferror(fp)) { result.clear(); return result; }
                if(n <= 0) break;
                result.append(Buf, Buf+n);
            }
            fclose(fp);
        }
        return result;
    }
    
    void SplitDecodeQuery(std::string& request,
                          std::multimap<std::string, std::string>& queryparams)
    {
        //fprintf(stderr, "Request(%s)\n", request.c_str());
        
        std::string::size_type p = request.find('?');
        if(p != request.npos)
        {
            std::string query = request.substr(p+1);
            request.erase(p);
            
            //fprintf(stderr, "Query(%s)\n", query.c_str());
            
            for(p = 0; p < query.size(); )
            {
                std::string::size_type nextp = query.find('&', p);
                if(nextp == query.npos) nextp = query.size();
                
                std::string token = query.substr(p, nextp-p);
                
                //fprintf(stderr, "  Token(%s)\n", token.c_str());
                
                std::string::size_type eqpos = token.find('='), eqaftpos = eqpos+1;
                if(eqpos == token.npos) eqpos = token.size(), eqaftpos = 0;
                
                std::string key   = urldecode(token.substr(0, eqpos));
                std::string value = urldecode(token.substr(eqaftpos));
                
                //fprintf(stderr, "    Key(%s)\n", key.c_str());
                //fprintf(stderr, "    Value(%s)\n", value.c_str());
                queryparams.insert(std::make_pair(key, value));
                
                p = nextp+1;
            }
        }
        
        // finally, unescape request
        request = urldecode(request);
    }
};

int main(void)
{
    std::signal(SIGPIPE, SIG_IGN);
    
    Tracker track;
    track.MainLoop();
}
