#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>

#include <errno.h>
#include <fcntl.h>

#include <stdio.h> // perror
#include <unistd.h> // close
#include <string.h> // memset

#include <sstream>

#include "socket.hh"

static std::string PeerIp;

int SockAccept(int fd, bool blocking)
{
    struct sockaddr_in addr;
    socklen_t addrlen = sizeof(addr);
    int sfd;

    if ((sfd = accept(fd, (struct sockaddr *)&addr, &addrlen)) == -1) {
        if (errno == EAGAIN || errno == EWOULDBLOCK) {
            return -1;
        } else {
            perror("accept()");
        }
        return -1;
    }
    
    if(!blocking)
    {
        int flags = -1;
        if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
            fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0) {
            perror("setting O_NONBLOCK");
            close(sfd);
            return -1;
        }
    }
    
    PeerIp = inet_ntoa(addr.sin_addr);
    
    return sfd;
}

const std::string& GetLastPeerIp()
{
    return PeerIp;
}

int SockAcceptNonblock(int fd)
{
    return SockAccept(fd, false);
}

int SockAcceptDoblock(int fd)
{
    return SockAccept(fd, true);
}

static int new_socket(bool blocking)
{
    int sfd;

    if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket()");
        return -1;
    }

    if(!blocking)
    {
        int flags;
        if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
            fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0) {
            perror("setting O_NONBLOCK");
            close(sfd);
            return -1;
        }
    }
    return sfd;
}

static int MakeSocket
    (int port, bool blocking,
     const char* address,
     bool do_bind)
{
    struct addrinfo ai, *aitop = 0;
    memset(&ai, 0, sizeof(ai));
    ai.ai_family   = AF_INET;
    ai.ai_socktype = SOCK_STREAM;
    ai.ai_flags    = do_bind ? AI_PASSIVE : 0;

    {std::stringstream strport; strport << port;
    if(getaddrinfo(address, strport.str().c_str(), &ai, &aitop) != 0)
    {
        perror("getaddrinfo()");
        return -1;
    }}

    int sfd;
    if ((sfd = new_socket(blocking)) == -1)
    {
        freeaddrinfo(aitop);
        return -1;
    }

    int flags=1;
    struct linger ling = {0, 0};
    setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
    setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags));
    setsockopt(sfd, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));
#if !defined(TCP_NOPUSH)
    setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(flags));
#endif
    
    if(do_bind)
    {
        if (bind(sfd, aitop->ai_addr, aitop->ai_addrlen) == -1) {
            perror("bind()");
        ErrOut:
            close(sfd);
            freeaddrinfo(aitop);
            return -1;
        }
        if (listen(sfd, 64) == -1) {
            perror("listen()");
            goto ErrOut;
        }
    }
    else
    {
        if(connect(sfd, aitop->ai_addr, aitop->ai_addrlen) == -1)
        {
            if(errno != EINPROGRESS)
            {
                perror("connect()");
                goto ErrOut;
            }
        }
    }
    
    freeaddrinfo(aitop);
    return sfd;
}

int MakeServerSocket(int port, bool blocking, const std::string& bind_host)
{
    return MakeSocket(port, blocking, bind_host.c_str(), true);
}

int MakeClientSocket(const std::string& host, int port, bool blocking)
{
    return MakeSocket(port, blocking, host.c_str(), false);
}
